#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "models.h"
#include "util.h"

void print_help(klist *ctx, char **argv);
void setup(klist *ctx, int argc, char **argv);

int main(int argc, char **argv) {
  klist *ctx = klist_init("test.db");

  if (argc == 1) {
    print_help(ctx, argv);
    ctx->error = 1;
  }

  if (!ctx->error) {
    setup(ctx, argc, argv);
    klist_user *user = klist_user_get_by_local(ctx, getuid());
    klist_list *list = NULL;
    int i = 0;
    switch (ctx->cmd) {
    case USER:
      klist_user_context *user_ctx = ctx->cmd_ctx;
      switch (user_ctx->cmd) {
      case USER_GET:
        if (user) {
          printf("User: %p\nID: %lu\n", user->name, user->id);
          size_t lists_len = 0;
          klist_list **lists =
              klist_list_get_all_by_user(ctx, user->id, &lists_len);
          printf("Lists: %lu\n", lists_len);
          for (; i < lists_len; i++)
            klist_list_deinit(lists[i]);
        } else
          klist_print(ctx, KLIST_LOG_ERROR, "No user for '%s' found.\n", getlogin());
        break;
      case USER_CREATE:
        klist_print(ctx, KLIST_LOG_INFO, (klist_assure_user(ctx, getuid(), getlogin()))
                            ? "User created.\n"
                            : "User already exists.\n");
        break;
      case USER_DELETE:
        if (user) {
          klist_user_delete(ctx, user);
          printf("User deleted.\n");
        } else
          klist_print(ctx, KLIST_LOG_ERROR, "User not found, no changes done.\n");
        break;
      }
      klist_user_context_deinit(user_ctx);
      break;
    case LIST:
      klist_list_context *list_ctx = ctx->cmd_ctx;
      if (!list_ctx->name && list_ctx->cmd != LIST_GET) {
        klist_print(ctx, KLIST_LOG_ERROR, "Missing name.\n");
        break;
      }
      switch (list_ctx->cmd) {
      case LIST_ADD:
        if (!list_ctx->stages_len) {
          klist_print(ctx, KLIST_LOG_ERROR, "Missing stages.\n");
          break;
        }
        klist_assure_user(ctx, getuid(), getlogin());
        list = klist_list_get_by_user_and_name(ctx, user->id, list_ctx->name);
        if (!list)
          list = klist_list_init();
        list->name = (unsigned char *)strdup(list_ctx->name);
        if (list->desc)
          list->desc = (unsigned char *)strdup(list_ctx->desc);
        list->is_preset = false;
        klist_list_save(ctx, list, user);
        i = 0;
        for (; i < list_ctx->stages_len; i++) {
          klist_stage *stage = klist_stage_init();
          stage->name = (unsigned char *)strdup(list_ctx->stages[i]);
          stage->list_id = list->id;
          klist_stage_save(ctx, stage);
          klist_stage_deinit(stage);
        }
        break;
      case LIST_EDIT:
        klist_print(ctx, KLIST_LOG_ERROR, "Not implemented\n");
        break;
      case LIST_GET:
        size_t lists_len = 0;
        klist_list **lists = NULL;
        if (list_ctx->name) {
          klist_list *list =
              klist_list_get_by_user_and_name(ctx, user->id, list_ctx->name);
          if (list) {
            lists_len = 1;
            lists = malloc(lists_len * sizeof(klist_list));
            memcpy(lists[0], list, sizeof(klist_list));
            klist_list_deinit(list);
          };
        } else
          lists = klist_list_get_all_by_user(ctx, user->id, &lists_len);

        if (lists_len) {
          i = 0;
          for (; i < lists_len; i++) {
            klist_list *list = lists[i];
            size_t tasks_len = 0;
            klist_task **tasks =
                klist_task_get_for_list(ctx, list->id, &tasks_len);
            printf("Name: %s\n", (char *)list->name);
            printf("Description: %s\n", list->desc ? (char *)list->desc : "N/A");
            printf("Tasks: %ld\n", tasks_len);
            i = 0;
            for (; i < tasks_len; i++)
              klist_task_deinit(tasks[i]);
            free(tasks);
            klist_list_deinit(list);
          }
          free(lists);
        } else if (list_ctx->name)
          klist_print(ctx, KLIST_LOG_ERROR, "List '%s' not found.\n", list_ctx->name);
        else
          klist_print(ctx, KLIST_LOG_WARNING, "No lists found.\n");
        break;
      case LIST_DELETE:
        list = klist_list_get_by_user_and_name(ctx, user->id, list_ctx->name);
        if (list) {
          size_t tasks_len = 0;
          klist_task **tasks =
              klist_task_get_for_list(ctx, list->id, &tasks_len);
          i = 0;
          for (; i < tasks_len; i++) {
            klist_task_delete(ctx, tasks[i]);
            klist_task_deinit(tasks[i]);
          }
          free(tasks);
        }
      }
      if (list_ctx)
        klist_list_context_deinit(list_ctx);
      break;
    case TASK:
      klist_task_context *task_ctx = ctx->cmd_ctx;
      switch (((klist_task_context *)ctx->cmd_ctx)->cmd) {
      case TASK_ADD: // basically the same things happen and edit can create if
                     // nescessary
      case TASK_EDIT:
        klist_assure_user(ctx, getuid(), getlogin());
        list = klist_list_get_by_user_and_name(ctx, user->id, task_ctx->list);
        if (list) {
          size_t stages_len = 0;
          klist_stage **stages =
              klist_stage_get_all_for_list(ctx, list->id, &stages_len);
          if (!task_ctx->stage) {
            klist_print(ctx, KLIST_LOG_ERROR, "Stage missing/wrong, please pass one of: ");
            i = 0;
            for (; i < stages_len; i++)
              fprintf(ctx->log_target, "%s ", (char *)stages[i]->name);
            fprintf(ctx->log_target, "\n");
            break;
          }
          i = 0;
          ssize_t stage_id = -1;
          for (; i < stages_len; i++)
            if (strcmp((char *)stages[i]->name, task_ctx->stage) == 0)
              stage_id = stages[i]->id;

          if (stage_id == -1) {
            klist_print(ctx, KLIST_LOG_ERROR,
                    "Stage %s not found. Use one of: ", task_ctx->stage);
            i = 0;
            for (; i < stages_len; i++)
              fprintf(ctx->log_target, "%s ", (char *)stages[i]->name);
            fprintf(ctx->log_target, "\n");
          }
          i = 0;
          for (; i < stages_len; i++)
            klist_stage_deinit(stages[i]);
          free(stages);
          if (ctx->error)
            break;

          klist_task *task =
              klist_task_get_for_list_by_name(ctx, list->id, task_ctx->name);
          if (!task)
            task = klist_task_init();
          if (task->name)
            free(task->name);
          task->name = (unsigned char *)strdup(task_ctx->name);
          if (task->desc)
            free(task->desc);
          if (task_ctx->desc)
            task->desc = (unsigned char *)strdup(task_ctx->desc);
          task->stage_id = stage_id;
          klist_task_save(ctx, task);
          klist_task_deinit(task);
        } else
          klist_print(ctx, KLIST_LOG_ERROR, "List not found.\n");
        break;
      case TASK_GET:
        klist_assure_user(ctx, getuid(), getlogin());
        user = klist_user_get_by_local(ctx, getuid());
        list = klist_list_get_by_user_and_name(ctx, user->id, task_ctx->list);
        if (list) {
          if (task_ctx->name) {
            klist_task *task =
                klist_task_get_for_list_by_name(ctx, list->id, task_ctx->name);
            if (task) {
              printf("Name: %s\n", (char *)task->name);
              printf("Description: %s\n", task->desc ? (char *)task->desc : "N/A");
              klist_stage *stage = klist_stage_get_by_id(ctx, task->stage_id);
              printf("Stage: %s\n", (char *)stage->name);
              klist_stage_deinit(stage);
              klist_task_deinit(task);
            } else
              klist_print(ctx, KLIST_LOG_ERROR, "Task not found.\n");
          } else {
            size_t tasks_len = 0;
            klist_task **tasks =
                klist_task_get_for_list(ctx, list->id, &tasks_len);
            i = 0;
            for (; i < tasks_len; i++) {
              printf("Name: %s\n", (char *)tasks[i]->name);
              printf("Description: %s\n", tasks[i]->desc ? (char *)tasks[i]->desc : "N/A");
              klist_stage *stage =
                  klist_stage_get_by_id(ctx, tasks[i]->stage_id);
              printf("Stage: %s\n", (char *)stage->name);
              klist_stage_deinit(stage);
              klist_task_deinit(tasks[i]);
              if ((i + 1) < tasks_len)
                fprintf(ctx->log_target, "\n");
            }
          }
        } else
          klist_print(ctx, KLIST_LOG_ERROR, "list not found\n");
      case TASK_DELETE:
        klist_assure_user(ctx, getuid(), getlogin());
        user = klist_user_get_by_local(ctx, getuid());
        list = klist_list_get_by_user_and_name(ctx, user->id, task_ctx->list);
        if (list) {
          if (task_ctx) {
            klist_task *task =
                klist_task_get_for_list_by_name(ctx, list->id, task_ctx->name);
            if (task) {
              klist_print(ctx, KLIST_LOG_WARNING, "Deleting task '%s'.\n", (char *)task->name);
              klist_task_delete(ctx, task);
              klist_task_deinit(task);
            } else
              klist_print(ctx, KLIST_LOG_ERROR, "Task not found.\n");
          } else
            klist_print(ctx, KLIST_LOG_ERROR, "Task not found, only deletion by name is "
                            "currently supoprted.\n");
        } else
          klist_print(ctx, KLIST_LOG_ERROR, "Not implemented\n");
      }
      klist_task_context_deinit(task_ctx);
    default:;
    }
    if (list)
      klist_list_deinit(list);
    if (user)
      klist_user_deinit(user);
  }

  const int error = ctx->error;
  klist_deinit(ctx);
  return error;
}

void print_help(klist *ctx, char **argv) {
  printf("Usage: %s <action>\n", argv[0]);
}

void setup(klist *ctx, int argc, char **argv) {
  if (strcmp(argv[1], "user") == 0)
    ctx->cmd = USER;
  else if (strcmp(argv[1], "list") == 0)
    ctx->cmd = LIST;
  else if (strcmp(argv[1], "task") == 0)
    ctx->cmd = TASK;
  else
    print_help(ctx, argv);

  int opt;
  switch (ctx->cmd) {
  case USER:
    optind = 2;
    klist_user_context *user_ctx = klist_user_context_init();
    while ((opt = getopt(argc, argv, "cdg")) != -1)
      switch (opt) {
      case 'c':
        user_ctx->cmd = USER_CREATE;
        break;
      case 'd':
        user_ctx->cmd = USER_DELETE;
        break;
      case 'g':
        user_ctx->cmd = USER_GET;
        break;
      default:
        print_help(ctx, argv);
      }
    ctx->cmd_ctx = user_ctx;
    break;
  case LIST:
    klist_list_context *list_ctx = klist_list_context_init();
    ctx->cmd_ctx = list_ctx;
    if (argc < 3) {
      list_ctx->cmd = LIST_GET;
      break;
    }

    optind = 2;
    while ((opt = getopt(argc, argv, "aden:p:s:")) != -1)
      switch (opt) {
      case 'a':
        list_ctx->cmd = LIST_ADD;
        break;
      case 'e':
        list_ctx->cmd = LIST_EDIT;
        break;
      case 'd':
        list_ctx->cmd = LIST_DELETE;
        break;
      case 'p':
        list_ctx->preset = strdup(optarg);
        break;
      case 'n':
        list_ctx->name = strdup(optarg);
        break;
      case 's':
        char *stage = NULL;
        while ((stage = strsep(&optarg, ","))) {
          list_ctx->stages = realloc(list_ctx->stages,
                                     (++list_ctx->stages_len) * sizeof(char *));
          list_ctx->stages[list_ctx->stages_len - 1] = strdup(stage);
        }
        break;
      default:
        print_help(ctx, argv);
        klist_print(ctx, KLIST_LOG_ERROR,
                ""
                "list options:\n"
                "-a\t\tadd a list\n"
                "-e\t\tedit a list\n"
                "-d\t\tdelete a list\n"
                "-p <preset>\tdefine as preset or use existing one with name "
                "<preset>\n"
                "Not providing a parameter prints all lists and existing "
                "presets.\n\n");
      }
    break;
  case TASK:
    if (argc < 3) {
      klist_print(ctx, KLIST_LOG_ERROR, "Prints all lists.\n");
      break;
    }
    klist_task_context *task_ctx = klist_task_context_init(ctx);

    optind = 2;

    while ((opt = getopt(argc, argv, "aedl:n:s:")) != -1)
      switch (opt) {
        /*case 'a':
        task_ctx->cmd = TASK_ADD;
        break;
      case 'e':
        task_ctx->cmd = TASK_EDIT;
        break;*/
      case 'd':
        task_ctx->cmd = TASK_DELETE;
        break;
      case 'l':
        task_ctx->list = strdup(optarg);
        break;
      case 'n':
        task_ctx->name = strdup(optarg);
        break;
      case 's':
        task_ctx->stage = strdup(optarg);
        break;
      default:
        print_help(ctx, argv);
        klist_print(ctx, KLIST_LOG_ERROR,
                ""
                "task options:"
                "-a\tadd a task"
                "-e\tedit a task"
                "-d\tdelete a task"
                "Not providing a parameter prints all tasks of a list.");
      }

    if (task_ctx->cmd != TASK_DELETE)
      task_ctx->cmd = TASK_ADD;
    ctx->cmd_ctx = task_ctx;
    break;
  default:
    klist_print(ctx, KLIST_LOG_ERROR, "How did we land here?! Pls report argv[1] = %s\n",
            argv[1]);
    print_help(ctx, argv);
  }
}
